
import re
import lxml
import logging
from odoo.tests import tagged
from odoo.addons.saas_worker.tests.util import countries
from odoo.addons.website.tests import test_crawl as website_test_crawl

_logger = logging.getLogger(__name__)

@tagged('post_install', '-at_install', 'crawl')
class Crawler(website_test_crawl.Crawler):

    # URL matching the following pattern will be discarded by the crawl
    # as they are targeting production URL or pages that don't exists in
    # local env.
    blacklisted_patterns = [
        # prod. only datas
        "^/blog/(customer-reviews-)?6(/.*)?$",
        "^/blog/(odoo-news-)?5(/.*)?$",
        "^/blog/(business-hacks-)?1(/.*)$",
        "^/jobs/country/belgium-20/.*$",  # only for geoip country == BE

        # prod. only website redirects or link tracker
        "^/elearning$",  # redirects. to /slides
        "^/learn$",  # redirects to /slides
        "^/r/HRGuidebook$",
        "^/r/Scale_Up_Download_[A-Z]{2}$",
        "^/r/Implementation_methodology$",
        '^/book/developers$',  # direct appointment link for "Developer on Demand"

        # external redirects (nginx)
        "^/documentation",
        "^/page/legal$",
        "^/legal$",

        # missings pages
        "^/labodoo$",  # only for geoip country == BE
        "^/security$",
        "^/privacy$",
        "^/website-builder-showcase$",
        "^/odoo-[0-9]{1,2}(-[0-9])?-release-notes$",

        # others
        "^/shop/change_pricelist/",  # (will crash /pricing if last crawled is incompatible)

        # White Papers vs pages
        "^/page/odoo-vs-.*$",

        # TMP. upgrade 17.0
        # - odoo/odoo@4a890e1f92a665dd53f17ee57b4e7792c33da7db
        #   (website_event_track template using `is_youtube_chat_available`
        #    that is only defined in website_event_track_live)
        '^/event/[^/]+/track/',
        # - odoo.addons.website.tests.test_crawl: Anonymous Coward /groups/public-mailing-list-2
        #   odoo.addons.base.models.ir_qweb: Unknown directives or unused attributes: {'t-key'} in mail_group.message_attachments
        '^/groups/',

    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls._blacklisted_patterns_used = set()

        # Create a partner in Belgium for each grade to avoid /partners to
        # return with a 404
        be_country_id = cls.env.ref('base.be').id
        partner_defaults = {'country_id': be_country_id, 'is_published': True, 'is_company': True}
        cls.env["res.partner"].create([
            dict(partner_defaults, name='%s partner' % grade.name, grade_id=grade.id,
                 commission_plan_id=grade.default_commission_plan_id.id)
            for grade in cls.env["res.partner.grade"].search([('is_published', '=', True)])
        ])

    def setUp(self):
        # skip website_test_crawl.Crawler own `setUp` method as it will try to
        # create a partner with a grade but without a commission plan, raising
        # a ValidationError from `_check_commission_plan_if_grade()` constrain
        super(website_test_crawl.Crawler, self).setUp()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        # ensure blacklist patterns does not contains unused/deprecated patterns.
        unused_blacklisted_patterns = set(cls.blacklisted_patterns) - cls._blacklisted_patterns_used
        if unused_blacklisted_patterns:  # pragma: nocover
            _logger.warning(
                "No URL matched the following blacklisted pattern during any "
                "crawling test, please check if they are still relevant:\n%s",
                "\n".join(f"- {pattern}" for pattern in unused_blacklisted_patterns)
            )

    def crawl(self, url, seen=None, msg=''):
        if seen is None:
            seen = set()

        root_crawl = url == "/" and url not in seen

        for pattern in self.blacklisted_patterns:
            if re.match(pattern, url):
                # skip fetching and crawling the page
                self._blacklisted_patterns_used.add(pattern)
                return seen

        seen = super().crawl(url, seen=seen, msg=msg)

        if root_crawl:
            # force extra pages from sitemap; i.e public pages; to be crawled too
            # (some of them may not be referenced directly on the website)
            seen = self.crawl_sitemap("/sitemap.xml", seen, msg=msg)

        return seen

    def crawl_sitemap(self, url, seen, msg=''):
        if url in seen:
            return seen
        else:
            seen.add(url)

        base_url = self.base_url()
        r = self.url_open(url, allow_redirects=False)
        code = r.status_code
        self.assertTrue(200 <= code < 300, "%s Fetching %s returned error response (%d)" % (msg, url, code))

        sitemap_html = lxml.html.fromstring(r.content)
        for split_sitemap_loc in sitemap_html.xpath('//sitemapindex/sitemap/loc'):
            split_sitemap_url = split_sitemap_loc.text
            if split_sitemap_url.startswith(base_url):
                split_sitemap_url = split_sitemap_url[len(base_url):]
            self.crawl_sitemap(split_sitemap_url, seen=seen, msg=msg)

        for url_loc in sitemap_html.xpath('//urlset/url/loc'):
            url = url_loc.text
            if url.startswith(base_url):
                url = url[len(base_url):]
            self.crawl(url, seen=seen, msg=msg)

        return seen

    @countries('BE', None)
    def test_10_crawl_public(self):
        super().test_10_crawl_public()

    @countries('BE')
    def test_20_crawl_demo(self):
        super().test_20_crawl_demo()
